# -*- coding: utf-8 -*-
from pyexpat import model
import unittest

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression as LinearRegression2
from sklearn.linear_model import Lasso as Lasso2
from sklearn.linear_model import LogisticRegression as LogisticRegression2
from sklearn.linear_model import BayesianRidge as BayesianRidge2
from sklearn.linear_model import Ridge as Ridge2
from sklearn.linear_model import TweedieRegressor as TweedieRegressor2
from sklearn.linear_model import HuberRegressor as HuberRegressor2
from sklearn.linear_model import SGDRegressor as SGDRegressor2
from sklearn.linear_model import PassiveAggressiveRegressor as PassiveAggressiveRegressor2
from sklearn.linear_model import TheilSenRegressor as TheilSenRegressor2

from ultron.optimize.model.linearmodel import ConstLinearModel
from ultron.optimize.model.linearmodel import LinearRegression
from ultron.optimize.model.linearmodel import LassoRegression
from ultron.optimize.model.linearmodel import LogisticRegression
from ultron.optimize.model.linearmodel import BayesianRegression
from ultron.optimize.model.linearmodel import RidgeRegression
from ultron.optimize.model.linearmodel import HuberRegression
from ultron.optimize.model.linearmodel import SGDRegression
from ultron.optimize.model.linearmodel import TweedieRegression
from ultron.optimize.model.linearmodel import PassiveAggressiveRegression
from ultron.optimize.model.linearmodel import TheilSenRegression
from ultron.optimize.model.loader import load_model


class LinearModel(unittest.TestCase):

    def setUp(self):
        self.n = 3
        self.features = ['a', 'b', 'c']
        self.train_x = pd.DataFrame(np.random.randn(1000, self.n),
                                    columns=self.features)
        self.train_y = np.random.randn(1000)
        self.train_y_label = np.where(self.train_y > 0., 1, 0)
        self.predict_x = pd.DataFrame(np.random.randn(10, self.n),
                                      columns=self.features)

    def test_const_linear_model(self):
        features = ['c', 'b', 'a']
        weights = dict(c=3., b=2., a=1.)
        model = ConstLinearModel(features=features, weights=weights)
        calculated_y = model.predict(self.predict_x)
        expected_y = self.predict_x[features] @ np.array(
            [weights[f] for f in features])
        np.testing.assert_array_almost_equal(calculated_y, expected_y)

    def test_const_linear_model_persistence(self):
        weights = dict(c=3., b=2., a=1.)
        model = ConstLinearModel(features=['a', 'b', 'c'], weights=weights)

        desc = model.save()
        new_model = load_model(desc)

        self.assertEqual(model.features, new_model.features)
        np.testing.assert_array_almost_equal(model.weights, new_model.weights)

    def test_const_linear_model_score(self):
        model = LinearRegression(['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        expected_score = model.score(self.train_x, self.train_y)

        const_model = ConstLinearModel(features=['a', 'b', 'c'],
                                       weights=dict(
                                           zip(model.features, model.weights)))

        calculated_score = const_model.score(self.train_x, self.train_y)

        self.assertAlmostEqual(expected_score, calculated_score)

    def test_linear_regression_persistence(self):
        model = LinearRegression(['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)

    def test_logistic_regression(self):
        model = LogisticRegression(['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y_label)

        calculated_y = model.predict(self.predict_x)

        expected_model = LogisticRegression2(fit_intercept=False)
        expected_model.fit(self.train_x, self.train_y_label)
        expected_y = expected_model.predict(self.predict_x)

        np.testing.assert_array_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(expected_model.coef_,
                                             model.weights)

    def test_logistic_regression_persistence(self):
        model = LogisticRegression(['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y_label)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)

    def test_bayesian_regression(self):
        model = BayesianRegression(features=['a', 'b', 'c'],
                                   fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        calculated_y = model.predict(self.predict_x)

        expected_model = BayesianRidge2(alpha_1=0.01, fit_intercept=False)
        expected_model.fit(self.train_x, self.train_y)
        expected_y = expected_model.predict(self.predict_x)

        np.testing.assert_array_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(expected_model.coef_,
                                             model.weights)

    def test_bayesian_regression_persistence(self):
        model = BayesianRegression(features=['a', 'b', 'c'],
                                   fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)

    def test_rigde_regression(self):
        model = RidgeRegression(features=['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        calculated_y = model.predict(self.predict_x)

        expected_model = Ridge2(alpha=0.01, fit_intercept=False)
        expected_model.fit(self.train_x, self.train_y)
        expected_y = expected_model.predict(self.predict_x)

        np.testing.assert_array_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(expected_model.coef_,
                                             model.weights)

    def test_rigde_regression_persistence(self):
        model = RidgeRegression(features=['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)

    def test_lasso_regression(self):
        model = LassoRegression(features=['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        calculated_y = model.predict(self.predict_x)

        expected_model = Lasso2(alpha=0.01, fit_intercept=False)
        expected_model.fit(self.train_x, self.train_y)
        expected_y = expected_model.predict(self.predict_x)

        np.testing.assert_array_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(expected_model.coef_,
                                             model.weights)

    def test_lasso_regression_persistence(self):
        model = LassoRegression(features=['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)

    def test_huber_regression(self):
        model = HuberRegression(features=['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        calculated_y = model.predict(self.predict_x)

        expected_model = HuberRegressor2(alpha=0.01, fit_intercept=False)
        expected_model.fit(self.train_x, self.train_y)
        expected_y = expected_model.predict(self.predict_x)

        np.testing.assert_array_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(expected_model.coef_,
                                             model.weights)

    def test_huber_regression_persistence(self):
        model = HuberRegression(features=['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)

    def test_sgd_regression_persistence(self):
        model = SGDRegression(features=['a', 'b', 'c'], fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)

    def test_tweedie_regression(self):
        model = TweedieRegression(features=['a', 'b', 'c'],
                                  fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        calculated_y = model.predict(self.predict_x)

        expected_model = TweedieRegressor2(alpha=0.01, fit_intercept=False)
        expected_model.fit(self.train_x, self.train_y)
        expected_y = expected_model.predict(self.predict_x)

        np.testing.assert_array_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(expected_model.coef_,
                                             model.weights)

    def test_tweedie_regression_persistence(self):
        model = TweedieRegression(features=['a', 'b', 'c'],
                                  fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)

    def test_pa_regression_persistence(self):
        model = PassiveAggressiveRegression(features=['a', 'b', 'c'],
                                            fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)

    def test_ths_regression_persistence(self):
        model = TheilSenRegression(features=['a', 'b', 'c'],
                                   fit_intercept=False)
        model.fit(self.train_x, self.train_y)

        desc = model.save()
        new_model = load_model(desc)

        calculated_y = new_model.predict(self.predict_x)
        expected_y = model.predict(self.predict_x)

        np.testing.assert_array_almost_equal(calculated_y, expected_y)
        np.testing.assert_array_almost_equal(new_model.weights, model.weights)



# -*- encoding:utf-8 -*-
"""示例ump主裁特征跳空模块"""
from ultron.ump.trade.ml_feature import FeatureJump
from ultron.ump.ump.base import ump_main_make_xy, BuyUmpMixin
from ultron.ump.ump.main_base import MainBase
from ultron.ump.model.principles import Principles


class JumpFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

        在__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_main_make_xy
    def make_xy(self, **kwarg):
        """
        make_xy通过装饰器ump_main_make_xy进行二次包装
        这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
        :param kwarg: ump_main_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                      kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

        :return: self.order_has_ret中使用filter选取特征列形成jump_df
                 ump_main_make_xy装饰器在make_xy返回jump_df后做转换matrix，形成x，y等工作
        """
        # 'regex=result|buy_diff_down_days|buy_diff_up_days|buy_jump_down_power|buy_jump_up_power'
        regex = 'result|{}'.format('|'.join(
            FeatureJump().get_feature_ump_keys(ump_cls=MainJump)))
        # noinspection PyUnresolvedReferences
        jump_df = self.order_has_ret.filter(regex=regex)
        """
            eg: jump_df
                        result  buy_jump_down_power  buy_diff_down_days  \
            2014-09-24       0                0.000                   0
            2014-10-24       0                0.000                   0
            2014-10-29       1               -1.109                 278
            2014-10-29       1                0.000                   0
            2014-10-29       1               -1.522                  75
            2014-10-29       1               -1.109                 278
            2014-11-03       1               -1.451                  78
            2014-11-11       1                0.000                   0
            2014-11-12       0               -1.863                  82
            2014-11-26       0               -2.115                  96
            ...            ...                  ...                 ...
            2016-03-14       1               -2.965                  56
            2016-03-14       0               -2.323                  67
            2016-03-30       0               -1.744                  74
            2016-04-04       1               -2.085                  36
            2016-04-13       0               -2.940                  47
            2016-04-14       0               -1.764                 100
            2016-04-15       0               -1.455                 101
            2016-04-15       0               -1.455                 101
            2016-04-29       1               -1.178                  24
            2016-04-29       1               -1.178                  24

                        buy_jump_up_power  buy_diff_up_days
            2014-09-24              3.344                61
            2014-10-24              3.344                91
            2014-10-29              2.920                95
            2014-10-29              1.283               243
            2014-10-29              3.727               238
            2014-10-29              2.920                95
            2014-11-03              1.014               214
            2014-11-11              1.291               256
            2014-11-12              1.606               256
            2014-11-26              1.772               270
            ...                       ...               ...
            2016-03-14              2.682                14
            2016-03-14              2.478               246
            2016-03-30              2.621                32
            2016-04-04              1.507                45
            2016-04-13              1.678                56
            2016-04-14              1.158                71
            2016-04-15              1.075                58
            2016-04-15              1.075                58
            2016-04-29              3.259                 0
            2016-04-29              3.259                 0
        """
        return jump_df


class MainJump(MainBase, BuyUmpMixin):

    def get_predict_col(self):
        """
        主裁跳空特征keys：['buy_diff_down_days', 'buy_diff_up_days', 'buy_jump_down_power', 'buy_jump_up_power']
        :return: ['buy_diff_down_days', 'buy_diff_up_days', 'buy_jump_down_power', 'buy_jump_up_power']
        """
        return FeatureJump().get_feature_ump_keys(ump_cls=MainJump)

    def get_fiter_class(self):
        """
        主裁特征跳空返回的UltronMLPd子类：UltronUmpMainJump.UmpJumpFiter
        :return: UltronUmpMainJump.UmpJumpFiter
        """
        return JumpFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'jump_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'jump_main'


# -*- coding: utf-8 -*-

import numpy as np
import time, datetime, pickle, itertools, os, copy
from joblib import Parallel, delayed
from .program import Program
from .operators import operators_sets
from ....utilities.jobs import partition_estimators
from ....utilities.utils import check_random_state
from ....utilities.mlog import MLog

import warnings

warnings.filterwarnings("ignore")

MAX_INT = np.iinfo(np.int32).max
MIN_INT = np.iinfo(np.int32).min


def parallel_evolve(n_programs, parents, total_data, seeds, greater_is_better,
                    gen, params):
    tournament_size = params['tournament_size']
    function_set = params['function_set']
    operators_set = params['operators_set']
    init_depth = params['init_depth']
    init_method = params['init_method']
    method_probs = params['method_probs']
    p_point_replace = params['p_point_replace']
    factor_sets = params['factor_sets']
    fitness = params['fitness']
    backup_cycle = params['backup_cycle']
    custom_params = params['custom_params']

    def _contenders(tour_parents):
        contenders = random_state.randint(0, len(tour_parents), 2)
        return [tour_parents[p] for p in contenders]

    def _tournament(tour_parents):
        contenders = random_state.randint(0, len(tour_parents),
                                          tournament_size)
        raw_fitness = [tour_parents[p]._raw_fitness for p in contenders]
        if greater_is_better:
            parent_index = contenders[np.argmax(raw_fitness)]
        else:
            parent_index = contenders[np.argmin(raw_fitness)]
        return tour_parents[parent_index], parent_index

    programs = []
    for i in range(n_programs):
        random_state = check_random_state(seeds[i])
        if parents is None:
            program = None
            genome = None
        else:
            method = random_state.uniform()
            parent, parent_index = _tournament(copy.deepcopy(parents))
            ori_parent = copy.deepcopy(parent)

            contenders = _contenders(copy.deepcopy(parents))
            for contender in contenders:
                program, removed, remains = parent.crossover(
                    contender._program, random_state)
                parent = Program(init_depth=init_depth,
                                 method=init_method,
                                 random_state=random_state,
                                 factor_sets=factor_sets,
                                 function_set=function_set,
                                 operators_set=operators_set,
                                 gen=gen,
                                 p_point_replace=p_point_replace,
                                 fitness=params['fitness'],
                                 n_features=2,
                                 program=program)

            #新特征种群加入
            if random_state.uniform() < method_probs[2]:
                program = Program(init_depth=init_depth,
                                  method=init_method,
                                  random_state=random_state,
                                  factor_sets=factor_sets,
                                  function_set=function_set,
                                  operators_set=operators_set,
                                  gen=gen,
                                  p_point_replace=p_point_replace,
                                  fitness=params['fitness'],
                                  n_features=2,
                                  program=None)
                program, removed, remains = parent.crossover(
                    program._program, random_state)

                parent = Program(init_depth=init_depth,
                                 method=init_method,
                                 random_state=random_state,
                                 factor_sets=factor_sets,
                                 function_set=function_set,
                                 operators_set=operators_set,
                                 gen=gen,
                                 p_point_replace=p_point_replace,
                                 fitness=params['fitness'],
                                 n_features=2,
                                 program=program)

            if method < method_probs[0]:  # # crossover
                donor, donor_index = _tournament(copy.deepcopy(parents))
                program, removed, remains = parent.crossover(
                    donor._program, random_state)
                genome = {
                    'method': 'Crossover',
                    'parent_idx': parent_index,
                    'parent_nodes': removed,
                    'donor_idx': donor_index,
                    'donor_nodes': remains
                }
            elif method < method_probs[1]:  # subtree_mutation
                program, removed, _ = parent.subtree_mutation(random_state)
                genome = {
                    'method': 'Subtree Mutation',
                    'parent_idx': parent_index,
                    'parent_nodes': removed
                }
            elif method < method_probs[2]:  # hoist_mutation
                program, removed = parent.hoist_mutation(random_state)
                genome = {
                    'method': 'Hoist Mutation',
                    'parent_idx': parent_index,
                    'parent_nodes': removed
                }
            elif method < method_probs[3]:  # point_mutation
                program, mutated = parent.point_mutation(random_state)
                genome = {
                    'method': 'Point Mutation',
                    'parent_idx': parent_index,
                    'parent_nodes': mutated
                }
            else:
                program = parent.reproduce()  # reproduction
                genome = {
                    'method': 'Reproduction',
                    'parent_idx': parent_index,
                    'parent_nodes': []
                }

            # 与原始自身进行交叉
            if random_state.uniform() < method_probs[3]:
                program = Program(init_depth=init_depth,
                                  method=init_method,
                                  random_state=random_state,
                                  factor_sets=factor_sets,
                                  function_set=function_set,
                                  operators_set=operators_set,
                                  gen=gen,
                                  p_point_replace=p_point_replace,
                                  fitness=params['fitness'],
                                  n_features=2,
                                  program=program,
                                  parents=genome)
                program, removed, remains = program.crossover(
                    ori_parent._program, random_state)

        program = Program(init_depth=init_depth,
                          method=init_method,
                          random_state=random_state,
                          factor_sets=factor_sets,
                          function_set=function_set,
                          operators_set=operators_set,
                          gen=gen,
                          p_point_replace=p_point_replace,
                          fitness=params['fitness'],
                          n_features=2,
                          program=program,
                          parents=genome)
        default_value = MIN_INT if greater_is_better else MAX_INT
        program.raw_fitness(total_data,
                            factor_sets,
                            default_value=default_value,
                            backup_cycle=backup_cycle,
                            custom_params=custom_params)

        programs.append(program)
    return programs


class Gentic(object):

    def __init__(
            self,
            population_size=2000,
            generations=MAX_INT,
            tournament_size=20,
            stopping_criteria=0.0,
            factor_sets=None,
            init_depth=(5, 6),
            init_method='full',
            operators_set=operators_sets,
            n_jobs=1,
            p_crossover=0.9,
            p_subtree_mutation=0.01,
            p_hoist_mutation=0.01,
            p_point_mutation=0.01,
            p_point_replace=0.05,
            greater_is_better=True,  #True 倒序， False 正序
            verbose=1,
            is_save=1,
            rootid=0,
            session=0,
            standard_score=2,  # None代表 根据tournament_size保留种群  standard_score保留种群
            out_dir='result',
            backup_cycle=0,  # 后备数据周期，主要用于在时间序列上的问题
            convergence=None,  # 收敛值，若为None，则不需要收敛值。
            low_memory=False,
            fitness=None,
            random_state=None,
            custom_params=None,
            save_model=None):
        self._population_size = population_size
        self._generations = MAX_INT if generations == 0 else generations
        self._tournament_size = tournament_size
        self._stopping_criteria = stopping_criteria
        self._factor_sets = factor_sets
        self._init_depth = init_depth
        self._init_method = init_method
        self._operators_set = operators_set
        self._function_set = [op.name for op in self._operators_set]
        self._p_crossover = p_crossover
        self._p_subtree_mutation = p_subtree_mutation
        self._p_hoist_mutation = p_hoist_mutation
        self._p_point_mutation = p_point_mutation
        self._p_point_replace = p_point_replace
        self._random_state = random_state
        self._greater_is_better = greater_is_better
        self._standard_score = standard_score
        self._fitness = fitness
        self._n_jobs = n_jobs
        self._backup_cycle = backup_cycle
        self._custom_params = custom_params
        self._low_memory = low_memory
        self._verbose = verbose
        self._is_save = is_save
        self._out_dir = out_dir
        self._convergence = convergence
        self._rootid = int(
            time.time() * 1000000 +
            datetime.datetime.now().microsecond) if rootid == 0 else rootid
        self._session = int(
            time.time() * 1000000 +
            datetime.datetime.now().microsecond) if session == 0 else session
        self._save_model = self.save_model if save_model is None else save_model
        self._con_time = 0
        self._best_fitness = 0

    def save_model(self, gen, rootid, session, best_programs):
        result_list = [{
            'transform': program.transform(),
            'fitness': program._raw_fitness
        } for program in best_programs]
        out_dir = os.path.join(self._out_dir, str(session))
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        filename = os.path.join(out_dir, 'ultron_' + str(gen) + '.pkl')
        with open(filename, 'wb') as f:
            pickle.dump([result_list], f)

    def filter_programs(self, gen, population):
        ## 保留符合条件的种群(1.种群有效 2.分数优于基准分 3.符合指定个数)
        valid_prorams = np.array(population)[[
            program._is_valid for program in population
        ]]  # 只保留有效种群

        ## 删除重复种群
        identification_dict = {}
        for program in valid_prorams:
            identification_dict[program._identification] = program

        valid_prorams = list(identification_dict.values())
        fitness = [program._raw_fitness for program in valid_prorams]
        if self._standard_score is not None:  #分数筛选且第二代开始
            if self._greater_is_better:
                best_programs = np.array([
                    program for program in valid_prorams
                    if program._raw_fitness > self._standard_score
                ])
            else:
                best_programs = np.array([
                    program for program in valid_prorams
                    if program._raw_fitness < self._standard_score
                ])

        #若不满足分数，则进行排序选出前_tournament_size
        if len(best_programs
               ) < self._tournament_size or self._standard_score is None:
            if self._greater_is_better:
                best_programs = np.array(valid_prorams)[np.argsort(
                    fitness)[-self._tournament_size:]]
            else:
                best_programs = np.array(valid_prorams)[np.argsort(
                    fitness)[:self._tournament_size]]
        return best_programs

    def train(self, total_data):
        random_state = check_random_state(self._random_state)
        self._method_probs = np.array([
            self._p_crossover, self._p_subtree_mutation,
            self._p_hoist_mutation, self._p_point_mutation
        ])

        self._method_probs = np.cumsum(self._method_probs)

        if self._method_probs[-1] > 1:
            raise ValueError('The sum of p_crossover, p_subtree_mutation, '
                             'p_hoist_mutation and p_point_mutation should '
                             'total to 1.0 or less.')

        if self._init_method not in ('half and half', 'grow', 'full'):
            raise ValueError('Valid program initializations methods include '
                             '"grow", "full" and "half and half". Given %s.' %
                             self.init_method)

        if (isinstance(self._init_depth, tuple)
                and len(self._init_depth) != 2):
            raise ValueError('init_depth should be a tuple with length two.')

        if (isinstance(self._init_depth, tuple)
                and (self._init_depth[0] > self._init_depth[1])):
            raise ValueError('init_depth should be in increasing numerical '
                             'order: (min_depth, max_depth).')

        params = {}
        params['tournament_size'] = self._tournament_size
        params['function_set'] = self._function_set
        params['operators_set'] = self._operators_set
        params['init_depth'] = self._init_depth
        params['init_method'] = self._init_method
        params['method_probs'] = self._method_probs
        params['p_point_replace'] = self._p_point_replace
        params['factor_sets'] = self._factor_sets
        params['fitness'] = self._fitness
        params['backup_cycle'] = self._backup_cycle
        params['custom_params'] = self._custom_params

        self._programs = []
        self._best_programs = None
        self._run_details = {
            'generation': [],
            'average_fitness': [],
            'best_fitness': [],
            'generation_time': [],
            'best_programs': []
        }

        prior_generations = len(self._programs)
        n_more_generations = self._generations - prior_generations
        for gen in range(prior_generations, self._generations):
            start_time = time.time()
            if gen == 0:
                parents = None
            else:
                parents = self._programs[gen - 1]
                parents = [parent for parent in parents if parent._is_valid]

            n_jobs, n_programs, starts = partition_estimators(
                self._population_size, self._n_jobs)

            seeds = random_state.randint(MAX_INT, size=self._population_size)
            population = Parallel(n_jobs=n_jobs, verbose=self._verbose)(
                delayed(
                    parallel_evolve)(n_programs[i], parents, total_data, seeds,
                                     self._greater_is_better, gen, params)
                for i in range(n_jobs))

            population = list(itertools.chain.from_iterable(population))

            #剔除无效因子
            population = [
                program for program in population if program._is_valid
            ]
            if len(population) == 0:
                break

            if self._best_programs is None:
                self._programs.append(population)
            else:
                identification_dict = {}
                valid_prorams = list(
                    np.concatenate([population, self._best_programs]))
                for program in valid_prorams:
                    identification_dict[program._identification] = program
                valid_prorams = list(identification_dict.values())
                self._programs.append(valid_prorams)
            '''
            if not self._low_memory:
                for old_gen in np.arange(gen, 0, -1):
                    indices = []
                    for program in self._programs[old_gen]:
                        if program is not None:
                            if 'parent_idx' in program._parents:
                                indices.append(program._parents['parent_idx'])
                    indices = set(indices)
                    population_size = len(population)
                    for idx in range(population_size):
                        if idx not in indices:
                            self._programs[old_gen - 1][idx] = None 
            elif gen > 0:
                self._programs[gen - 1] = None
            '''

            best_programs = self.filter_programs(gen, population)

            if self._best_programs is not None:
                best_programs = np.concatenate(
                    [best_programs, self._best_programs])
                best_programs = self.filter_programs(gen, best_programs)

            self._best_programs = best_programs
            for program in self._best_programs:
                program.log()
            fitness = [program._raw_fitness for program in self._best_programs]
            self._run_details['generation'].append(gen)
            self._run_details['average_fitness'].append(np.mean(fitness))
            generation_time = time.time() - start_time
            self._run_details['generation_time'].append(generation_time)
            self._run_details['best_programs'].append(self._best_programs)
            MLog().write().info(
                'ExpendTime:%f,Generation:%d,Tournament:%d, Fitness Mean:%f,Fitness Max:%f,Fitness Min:%f'
                % (generation_time, gen, len(best_programs), np.mean(fitness),
                   np.max(fitness), np.min(fitness)))

            #保存每代信息
            if self._is_save:
                self._save_model(gen, self._rootid, self._session,
                                 self._run_details['best_programs'][-1],
                                 self._custom_params)

            if self._greater_is_better:
                best_fitness = fitness[np.argmax(fitness)]
                if best_fitness >= self._stopping_criteria:
                    break
            else:
                best_fitness = fitness[np.argmin(fitness)]
                if best_fitness <= self._stopping_criteria:
                    break

            if np.mean(fitness) == MIN_INT or best_fitness == MIN_INT:
                break
            self._run_details['best_fitness'].append(best_fitness)

            # 收敛值判断
            if self._convergence is None or gen == 0:
                continue
            d_value = np.mean(fitness) - self._run_details['average_fitness'][
                gen - 1]
            print('d_value:%f,convergence:%f,con_time:%d' %
                  (d_value, self._convergence, self._con_time))
            if abs(d_value) < self._convergence:
                self._con_time += 1
                if self._con_time > 5:
                    break
            else:
                self._con_time = 0



# -*- coding: utf-8 -*-
import datetime
import json
import subprocess
from ultron.cluster.work.extern_modules.base_module import BaseModule


class Module(BaseModule):

    def __init__(self, name, wid, token, redis_client):
        super(Module, self).__init__(name, wid, token, redis_client)
        self._func = {
            'startup_task': self.startup_task,
            'shutoff_task': self.shutoff_task,
            'restart_task': self.restart_task,
            'update_task': self.update_task
        }
        self._namespace = 'tasks'

    def process_respone(self, respone):
        name = respone['name']
        opcode = respone['opcode']
        self._func[opcode](respone)

    def update_task(self, respone):
        dir_name = respone['dir_name']
        tasks = respone['tasks']
        for task in tasks:
            dir_line = 'cd ' + dir_name
            stop_line = 'ps auxww | grep \'active- (-A  ' + str(
                task) + ' worker\' | awk \'{print $2}\' | xargs kill -9'
            subprocess.Popen(stop_line, shell=True)
            celery_line = 'celery -A ' + task + ' worker --loglevel=info'
            command_line = dir_line + ' && ' + celery_line
            subprocess.Popen(command_line, shell=True)

    def startup_task(self, respone):
        work_name = respone['work_name']
        dir_name = respone['dir_name']
        #celery -A polymeriza worker --loglevel=info
        dir_line = 'cd ' + dir_name
        stop_line = 'ps auxww | grep \'active- (-A  ' + str(
            work_name) + ' worker\' | awk \'{print $2}\' | xargs kill -9'
        celery_line = 'celery -A ' + work_name + ' worker --loglevel=info -E'
        command_line = dir_line + ' && ' + celery_line
        subprocess.Popen(command_line, shell=True)

    def shutoff_task(self, respone):
        work_name = respone['work_name']
        dir_name = respone['dir_name']
        command_line = 'ps auxww | grep \'active- (-A  ' + str(
            work_name) + ' worker\' | awk \'{print $2}\' | xargs kill -9'
        subprocess.Popen(command_line, shell=True)

    def restart_task(self):
        work_name = respone['work_name']
        dir_name = respone['dir_name']
        dir_line = 'cd ' + dir_name
        kill_line = 'ps auxww | grep \'active- (-A  ' + str(
            work_name) + ' worker\' | awk \'{print $2}\' | xargs kill -9'
        start_line = 'celery -A ' + work_name + ' worker --loglevel=info'
        command_line = dir_line + ' && ' + kill_line + ' && ' + start_line
        subprocess.Popen(command_line, shell=True)



# coding=utf-8
import multiprocessing
import pandas as pd
import datetime
import sqlalchemy as sa
from sqlalchemy.pool import NullPool
from sqlalchemy import select, and_, func

class CalcPool(object):
    
    def __init__(self, is_debug=False):
        __str__ = 'calc_pool'
        self._cpus = multiprocessing.cpu_count()
        self._is_debug = is_debug
        
    def _calc_fetch_info(self, cond):
        conn = sa.create_engine(cond[0], poolclass=NullPool)
        callback = cond[1]
        cond = cond[2]
        return callback(conn, cond)
    
    def calc_work_db(self, db_info, callback, params_list):
        cond_list = []
        for param in params_list:
            cond_list.append([db_info, callback, param])
            
        with multiprocessing.Pool(processes=self._cpus) as p:
            res = p.map(self._calc_fetch_info, cond_list)

        return pd.concat(res).reset_index(drop=True)
    
    def calc_work_method(self, callback, params_list, is_pandas=True):
        with multiprocessing.Pool(processes=self._cpus) as p:
            res = p.map(callback, params_list)
        return pd.concat(res).reset_index(drop=True) if is_pandas else res


# -*- encoding:utf-8 -*-
"""
    买入卖出因子与ump进行组织管理通信模块
"""

from ultron.kdutils.decorator import first_delegate_has_method, replace_word_delegate_has_method
from ultron.kdutils.file import file_exist
from ultron.ump.core.fixes import six
from ultron.ump.core import env
from ultron.utilities.lazy import LazyFunc
from ultron.ump.ump.edge_base import EEdgeType
from ultron.ump.ump.main_deg import MainDeg
from ultron.ump.ump.main_jump import MainJump
from ultron.ump.ump.main_price import MainPrice
from ultron.ump.ump.main_wave import MainWave
from ultron.ump.ump.main_base import MainBase
from ultron.ump.ump.edge_deg import EdgeDeg
from ultron.ump.ump.edeg_price import EdgePrice
from ultron.ump.ump.edge_wave import EdgeWave
from ultron.ump.ump.edge_full import EdgeFull
"""外部用户使用append_user_ump添加到外部ump list容器中"""
_g_extend_ump_list = list()
"""是否启用外部用户使用append_user_ump添加的ump对交易进行拦截决策"""
g_enable_user_ump = False


def append_user_ump(ump, check=True):
    """
    外部用户设置扩展ump接口, 参数ump可以是ump class类型，也可以是实例化后的ump object
    :param ump: 可以是ump class类型，也可以是实例化后的ump object
    :param check: 是否检测ump是否以及训练好，且生成了cache文件
    """
    if check:
        # 检测ump训练后的本地物理文件是否存在
        if isinstance(ump, six.class_types):
            ump_cache_path = ump(predict=True).dump_file_fn()
        else:
            ump_cache_path = ump.dump_file_fn()
        if not file_exist(ump_cache_path):
            # 如果用户添加的ump没有经过训练集训练，提升用户进行训练
            raise RuntimeError(
                'you must first fit orders, {} is not exist!!'.format(
                    ump_cache_path))
    _g_extend_ump_list.append(ump)


def clear_user_ump():
    """
    将外部用户设置的ump队列清空
    :return:
    """
    global _g_extend_ump_list
    _g_extend_ump_list = list()


class Manager(object):
    """ump组织管理类"""

    def __init__(self, market_name, is_buy_factor=True):
        self.extend_ump_list = list()
        self.is_buy_factor = is_buy_factor
        if g_enable_user_ump and len(_g_extend_ump_list) > 0:
            # 筛选与因子对应的ump，即买入因子UltronFactorBuyBase对象只筛选买入BuyUmpMixin类型的ump
            filter_ump = list(
                filter(lambda ump: self.is_buy_factor == ump.is_buy_ump(),
                       _g_extend_ump_list))
            self.extend_ump_list.extend(filter_ump)
        self._fix_ump_env()
        self._market_name = market_name

    def __str__(self):
        """打印对象显示：class name, factor, extend_ump_list"""
        return '{}: self.extend_ump_list={}'.format(self.__class__.__name__,
                                                    self.extend_ump_list)

    __repr__ = __str__

    @classmethod
    def _fix_ump_env(cls):
        """
        只为修复在UltronEnv中对ump的设置和manger中的ump设置不同步问题，即UltronEnv.g_enable_ml_feature
        设置不正
        """
        if env.g_enable_ump_main_deg_block or env.g_enable_ump_main_jump_block or \
            env.g_enable_ump_main_price_block or env.g_enable_ump_main_price_block or \
                env.g_enable_ump_main_wave_block:
            env.g_enable_ml_feature = True
        if g_enable_user_ump:
            env.g_enable_ml_feature = True

    @LazyFunc
    def ump_main_deg(self):
        """
            根据UltronEnv.g_enable_ump_main_deg_block设置决定是否构造UltronUmpMainDeg，
            使用LazyFunc装饰器，构造后即使用缓存UltronUmpMainDeg(predict=True)实例
        """
        return MainDeg(predict=True, market_name=self._market_name)

    @LazyFunc
    def ump_main_jump(self):
        """
            根据UltronEnv.g_enable_ump_main_jump_block设置决定是否构造UltronUmpMainJump，
            使用LazyFunc装饰器，构造后即使用缓存UltronUmpMainJump(predict=True)实例
        """
        return MainJump(predict=True, market_name=self._market_name)

    @LazyFunc
    def ump_main_price(self):
        """
            根据UltronEnv.g_enable_ump_main_jump_block设置决定是否构造UltronUmpMainJump，
            使用LazyFunc装饰器，构造后即使用缓存UltronUmpMainJump(predict=True)实例
        """
        return MainPrice(predict=True, market_name=self._market_name)

    @LazyFunc
    def ump_main_wave(self):
        """
            根据UltronEnv.g_enable_ump_main_jump_block设置决定是否构造UltronUmpMainJump，
            使用LazyFunc装饰器，构造后即使用缓存UltronUmpMainJump(predict=True)实例
        """
        return MainWave(predict=True, market_name=self._market_name)

    @LazyFunc
    def ump_edge_deg(self):
        """
            根据UltronEnv.g_enable_ump_edge_deg_block设置决定是否构造UltronUmpEdgeDeg，
            使用LazyFunc装饰器，构造后即使用缓存UltronUmpEdgeDeg(predict=True)实例
        """
        return EdgeDeg(predict=True, market_name=self._market_name)

    @LazyFunc
    def ump_edge_price(self):
        """
            根据UltronEnv.g_enable_ump_edge_price_block设置决定是否构造UltronUmpEdgePrice，
            使用LazyFunc装饰器，构造后即使用缓存UltronUmpEdgePrice(predict=True)实例
        """
        return EdgePrice(predict=True, market_name=self._market_name)

    @LazyFunc
    def ump_edge_wave(self):
        """
            根据UltronEnv.g_enable_ump_edge_wave_block设置决定是否构造UltronUmpEdgeWave，
            使用LazyFunc装饰器，构造后即使用缓存UltronUmpEdgeWave(predict=True)实例
        """
        return EdgeWave(predict=True, market_name=self._market_name)

    @LazyFunc
    def ump_edge_full(self):
        """
            根据UltronEnv.g_enable_ump_edge_full_block设置决定是否构造UltronUmpEdgeFull，
            使用LazyFunc装饰器，构造后即使用缓存UltronUmpEdgeFull(predict=True)实例
        """
        return EdgeFull(predict=True, market_name=self._market_name)

    # noinspection PyMethodMayBeStatic
    def _default_main_hit_cnt(self):
        # noinspection PyProtectedMember
        default_hit_cnt = 1  #if UltronEnv._g_enable_example_env_ipython else 2
        return default_hit_cnt

    @first_delegate_has_method(delegate='factor')
    def ump_main_deg_hit_cnt(self):
        """
        角度主裁使用predict_kwargs的参数need_hit_cnt值，即:
                predict_kwargs(need_hit_cnt=self.ump_main_deg_hit_cnt(), **ml_feature_dict)

        被装饰器first_delegate_has_method(delegate='factor')装饰，当被委托的因子，即self.factor中
        有对应实现的ump_main_deg_hit_cnt方法时，返回self.factor.ump_main_deg_hit_cnt()的返回值
        :return: int
        """
        return self._default_main_hit_cnt()

    @first_delegate_has_method(delegate='factor')
    def ump_main_jump_hit_cnt(self):
        """
        跳空主裁使用predict_kwargs的参数need_hit_cnt值，即:
                predict_kwargs(need_hit_cnt=self.ump_main_jump_hit_cnt, **ml_feature_dict)

        被装饰器first_delegate_has_method(delegate='factor')装饰，当被委托的因子，即self.factor中
        有对应实现的ump_main_jump_hit_cnt方法时，返回self.factor.ump_main_jump_hit_cnt()的返回值
        :return: int
        """
        return self._default_main_hit_cnt()

    @first_delegate_has_method(delegate='factor')
    def ump_main_price_hit_cnt(self):
        """
        价格主裁使用predict_kwargs的参数need_hit_cnt值，即:
                predict_kwargs(need_hit_cnt=self.ump_main_price_hit_cnt, **ml_feature_dict)

        被装饰器first_delegate_has_method(delegate='factor')装饰，当被委托的因子，即self.factor中
        有对应实现的ump_main_price_hit_cnt方法时，返回self.factor.ump_main_price_hit_cnt()的返回值
        :return: int
        """
        return self._default_main_hit_cnt()

    @first_delegate_has_method(delegate='factor')
    def ump_main_wave_hit_cnt(self):
        """
        价格波动主裁使用predict_kwargs的参数need_hit_cnt值，即:
                predict_kwargs(need_hit_cnt=self.ump_main_wave_hit_cnt, **ml_feature_dict)

        被装饰器first_delegate_has_method(delegate='factor')装饰，当被委托的因子，即self.factor中
        有对应实现的ump_main_wave_hit_cnt方法时，返回self.factor.ump_main_wave_hit_cnt()的返回值
        :return: int
        """
        return self._default_main_hit_cnt()

    # noinspection PyMethodMayBeStatic
    def ump_main_user_hit_cnt(self):
        """
        用户自定义的主裁ump类的使用predict_kwargs的参数need_hit_cnt值，
        用户在因子中可通过实现特点的方法名称来替换ump_main_user_hit_cnt值，

        使用时使用replace_word_delegate_has_method检测self.factor中有没有对应的方法，
        eg:
            replace_hit_cnt = replace_word_delegate_has_method(delegate='factor', key_word='user',
                                                                       replace_word=class_unique_id)
            hit_cnt = replace_hit_cnt(self.ump_main_user_hit_cnt)()

            即如果用户编写的主裁ump中class_unique_id方法返回'extend_main_test'
                @classmethod
                def class_unique_id(cls):
                    return 'extend_main_test'
            则在因子中对应自定义hit_cnt的方法名称应为：
                def ump_main_extend_main_test_hit_cnt(self)
                    return 1
            更多具体实现阅extend_ump_block以及replace_word_delegate_has_method方法的实现

        :return: int
        """
        return self._default_main_hit_cnt()

    def ump_block(self, ml_feature_dict):
        """
        在买入或者卖出因子中make_ump_block_decision方法中使用，决策特定交易是否被拦截，
        ump_block中首先使用内置ump进行拦截决策，如果不被拦截，使用外部定义的ump进行拦截决策

        :param ml_feature_dict: 交易所形成的特征字典
                eg: ml_feature_dict
                    {'buy_deg_ang42': -0.45400000000000001, 'buy_deg_ang252': 5.532,
                    'buy_deg_ang60': 2.1419999999999999, 'buy_deg_ang21': 0.93100000000000005,
                    'buy_price_rank120': 1.0, 'buy_price_rank90': 1.0, 'buy_price_rank60': 1.0,
                    'buy_price_rank252': 1.0, 'buy_wave_score1': 1.2470000000000001, 'buy_wave_score2': 1.286,
                    'buy_wave_score3': 1.2849999999999999, 'buy_atr_std': 0.19400000000000001,
                    'buy_jump_down_power': -13.57, 'buy_diff_down_days': 136, 'buy_jump_up_power': 1.038,
                    'buy_diff_up_days': 2}
        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """
        # 内置ump进行拦截决策
        if self.builtin_ump_block(ml_feature_dict):
            return True

        # 外部定义的ump进行拦截决策
        if self.extend_ump_block(ml_feature_dict):
            return True
        return False

    def extend_ump_block(self, ml_feature_dict):
        """
        外部用户设置的ump进行拦截决策，迭代self.extend_ump_list中外部设置的ump，
        由于对外添加ump的接口append_user_ump中参数ump可以是ump class类型，
        也可以是实例化后的ump object，所以需要把class类型的ump进行实例构造，且将
        实例的ump对象缓存在类变量中（通过class_unique_id为类变量构造唯一名称），
        ump对象构造好后根据主裁还是边裁选择决策方法：
                主裁使用：predict_kwargs(need_hit_cnt=need_hit_cnt, **ml_feature_dict)
                边裁使用：predict(**ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS
        对交易进行拦截决策

        :param ml_feature_dict: 交易所形成的特征字典
        eg: ml_feature_dict
            {'buy_deg_ang42': -0.45400000000000001, 'buy_deg_ang252': 5.532,
            'buy_deg_ang60': 2.1419999999999999, 'buy_deg_ang21': 0.93100000000000005,
            'buy_price_rank120': 1.0, 'buy_price_rank90': 1.0, 'buy_price_rank60': 1.0,
            'buy_price_rank252': 1.0, 'buy_wave_score1': 1.2470000000000001, 'buy_wave_score2': 1.286,
            'buy_wave_score3': 1.2849999999999999, 'buy_atr_std': 0.19400000000000001,
            'buy_jump_down_power': -13.57, 'buy_diff_down_days': 136, 'buy_jump_up_power': 1.038,
            'buy_diff_up_days': 2}
        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """
        for extend_ump in self.extend_ump_list:
            class_unique_id = extend_ump.class_unique_id()
            # 由于对外添加ump的接口append_user_ump中参数ump可以是ump class类型，也可以是实例化后的ump object
            if isinstance(extend_ump, six.class_types):
                # 把class类型的ump进行实例构造

                is_main_ump = issubclass(extend_ump, MainBase)
                main_ump_key = 'main' if is_main_ump else 'edge'

                # 通过class_unique_id和issubclass(extend_ump, UltronUmpMainBase)为类变量构造唯一名称
                extend_ump_attr_str = 'ump_{}_{}'.format(
                    main_ump_key, class_unique_id)
                if hasattr(self, extend_ump_attr_str):
                    # 将类变量中的实例代替类
                    extend_ump = getattr(self, extend_ump_attr_str)
                else:
                    # 内置ump通过LazyFunc进行效率提升，外部设置的ump通过手动setattr，将实例的ump对象缓存在类变量中
                    extend_ump_obj = extend_ump(predict=True)
                    setattr(self, extend_ump_attr_str, extend_ump_obj)
                    # 将实例化后的实例代替类
                    extend_ump = extend_ump_obj
            is_main_ump = isinstance(extend_ump, MainBase)
            if is_main_ump:
                # replace_word_delegate_has_method不做装饰器修饰ump_main_user_hit_cnt，因为需要动态获取replace_word
                try:
                    replace_hit_cnt = replace_word_delegate_has_method(
                        delegate='factor',
                        key_word='user',
                        replace_word=class_unique_id)
                    hit_cnt = replace_hit_cnt(self.ump_main_user_hit_cnt)()
                except:
                    # 忽略用户自定义factor中关于hit_cnt的任何错误
                    hit_cnt = self.ump_main_user_hit_cnt()
                if extend_ump.predict_kwargs(need_hit_cnt=hit_cnt,
                                             **ml_feature_dict):
                    return True
            else:
                return False  # 边裁
        return False

    def builtin_ump_block(self, ml_feature_dict):
        """
        内置ump进行拦截决策，通过UltronEnv中的拦截设置以及因子的买入卖出类型是否和ump类型匹配，
        来决定是否使用特定的ump进行拦截决策，如需要决策：

                主裁使用：predict_kwargs(need_hit_cnt=need_hit_cnt, **ml_feature_dict)
                边裁使用：predict(**ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS
        对交易进行拦截决策

                :param ml_feature_dict: 交易所形成的特征字典
                eg: ml_feature_dict
                    {'buy_deg_ang42': -0.45400000000000001, 'buy_deg_ang252': 5.532,
                    'buy_deg_ang60': 2.1419999999999999, 'buy_deg_ang21': 0.93100000000000005,
                    'buy_price_rank120': 1.0, 'buy_price_rank90': 1.0, 'buy_price_rank60': 1.0,
                    'buy_price_rank252': 1.0, 'buy_wave_score1': 1.2470000000000001, 'buy_wave_score2': 1.286,
                    'buy_wave_score3': 1.2849999999999999, 'buy_atr_std': 0.19400000000000001,
                    'buy_jump_down_power': -13.57, 'buy_diff_down_days': 136, 'buy_jump_up_power': 1.038,
                    'buy_diff_up_days': 2}

        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """
        """内置主裁开始"""
        if env.g_enable_ump_main_deg_block and self.ump_main_deg.predict_kwargs(
                need_hit_cnt=self.ump_main_deg_hit_cnt(), **ml_feature_dict):
            return True

        if env.g_enable_ump_main_jump_block and self.ump_main_jump.predict_kwargs(
                need_hit_cnt=self.ump_main_jump_hit_cnt(), **ml_feature_dict):
            return True

        if env.g_enable_ump_main_price_block and self.ump_main_price.predict_kwargs(
                need_hit_cnt=self.ump_main_price_hit_cnt(), **ml_feature_dict):
            return True

        if env.g_enable_ump_main_wave_block and self.ump_main_wave.predict_kwargs(
                need_hit_cnt=self.ump_main_wave_hit_cnt(), **ml_feature_dict):
            return True

        if env.g_enable_ump_edge_deg_block and self.ump_edge_deg.predict(
                **ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS:
            return True

        if env.g_enable_ump_edge_deg_block and self.ump_edge_deg.predict(
                **ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS:
            return True

        if env.g_enable_ump_edge_price_block and self.ump_edge_price.predict(
                **ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS:
            return True

        if env.g_enable_ump_edge_full_block and self.ump_edge_full.predict(
                **ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS:
            return True

        return False



# -*- coding: utf-8 -*-
import six
import json
import os
import shutil
from ultron.utilities.singleton import Singleton
from ultron.setting import setting_info, file_name


@six.add_metaclass(Singleton)
class ConfigSingletion(object):

    def __init__(self):
        __str__ = 'ConfigSingletion'
        self.queue_host = ''
        self.queue_port = 0
        self.queue_pwd = ''
        self.queue_user = ''
        self.queue_db = 0
        self.queue_type = ''
        self._init_setting()

    def _init_setting(self):
        l = json.loads(setting_info)
        for setting in l:
            values = l[setting]
            if setting == 'queue':
                self._init_queue(setting, values)

    def _init_queue(self, key, kwargs):
        self.set_queue(kwargs)

    def set_queue(self, kwargs):
        self.queue_type = kwargs['type']
        if self.queue_type == 'redis':
            self.queue_host = kwargs['host']
            self.queue_port = kwargs['port']
            self.queue_pwd = kwargs['pwd']
            self.queue_db = kwargs['db'] if 'db' in kwargs else 0
        elif self.queue_type == 'amqp':
            self.queue_host = kwargs['host']
            self.queue_port = kwargs['port']
            self.queue_user = kwargs['user']
            self.queue_pwd = kwargs['pwd']
            self.queue_db = kwargs['db'] if 'db' in kwargs else 0

    def update(self):
        self._save_setting()

    def save_queue(self):
        kwargs = {}
        kwargs['type'] = self.queue_type
        kwargs['host'] = self.queue_host
        kwargs['port'] = self.queue_port
        kwargs['user'] = self.queue_user
        kwargs['pwd'] = self.queue_pwd
        kwargs['db'] = self.queue_db
        return kwargs

    def _save_setting(self):
        kwargs = {}
        kwargs['queue'] = self.save_queue()
        info_str = json.dumps(kwargs)
        content = "import os\nsetting_info='{0}'\nfile_name = os.path.join(os.getcwd(), __file__)".format(
            info_str)
        #卸载引用库
        pyc_path = str(file_name.split('.')[0]) + '.pyc'
        if os.path.exists(pyc_path):
            os.remove(pyc_path)
        if os.path.exists(file_name):
            os.remove(file_name)
        with open(file_name, 'wb') as f:
            f.write(bytes(content, encoding='utf8'))


config_setting = ConfigSingletion()


# -*- encoding:utf-8 -*-
"""示例ump单混特征模块"""

from ultron.ump.trade.ml_feature import FeatureDeg, FeaturePrice, FeatureJump, FeatureWave, FeatureAtr
from ultron.ump.ump.base import ump_main_make_xy, BuyUmpMixin
from ultron.ump.ump.main_base import MainBase
from ultron.ump.model.principles import Principles


class MulFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

        在__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_main_make_xy
    def make_xy(self, **kwarg):
        """
        make_xy通过装饰器ump_main_make_xy进行二次包装
        这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
        :param kwarg: ump_main_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                      kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

        :return: self.order_has_ret中使用filter选取特征列形成mul_df
                 ump_main_make_xy装饰器在make_xy返回mul_df后做转换matrix，形成x，y等工作
        """

        # regex='result|buy_deg_ang21|buy_price_rank252|buy_wave_score3|buy_atr_std'
        regex = 'result|{}|{}|{}|{}'.format(
            FeatureDeg().get_feature_ump_keys(ump_cls=MainMul)[-1],
            FeaturePrice().get_feature_ump_keys(ump_cls=MainMul)[-1],
            FeatureWave().get_feature_ump_keys(ump_cls=MainMul)[-1],
            FeatureAtr().get_feature_ump_keys(ump_cls=MainMul)[-1])
        # noinspection PyUnresolvedReferences
        mul_df = self.order_has_ret.filter(regex=regex)
        """
            eg: mul_df
                        result  buy_deg_ang21  buy_price_rank252  buy_wave_score3  \
            2014-09-24       0          1.818              1.000            0.218
            2014-10-24       0         -1.089              1.000            0.449
            2014-10-29       1          1.980              1.000            0.329
            2014-10-29       1          6.507              0.750           -0.203
            2014-10-29       1          7.046              0.982           -0.173
            2014-10-29       1          1.980              1.000            0.329
            2014-11-03       1         10.125              1.000           -0.172
            2014-11-11       1         12.397              0.808            0.001
            2014-11-12       0          6.671              0.560           -0.509
            2014-11-26       0         12.494              0.762            0.015
            ...            ...            ...                ...              ...
            2016-03-14       1          9.324              0.444            0.948
            2016-03-14       0          5.201              0.623            0.788
            2016-03-30       0          4.070              0.536            0.702
            2016-04-04       1          5.618              0.190            0.743
            2016-04-13       0          4.457              0.270            0.651
            2016-04-14       0          0.733              0.940            0.170
            2016-04-15       0          1.162              0.631            0.800
            2016-04-15       0          1.162              0.631            0.800
            2016-04-29       1         -1.115              1.000            0.918
            2016-04-29       1         -1.115              1.000            0.918

                        buy_atr_std
            2014-09-24        0.226
            2014-10-24        0.146
            2014-10-29        0.538
            2014-10-29        0.558
            2014-10-29        0.063
            2014-10-29        0.538
            2014-11-03       -0.002
            2014-11-11        0.171
            2014-11-12        0.093
            2014-11-26        0.812
            ...                 ...
            2016-03-14        0.126
            2016-03-14        0.663
            2016-03-30       -0.008
            2016-04-04        0.035
            2016-04-13        0.262
            2016-04-14        0.049
            2016-04-15        0.135
            2016-04-15        0.135
            2016-04-29        0.838
            2016-04-29        0.838
        """
        return mul_df


class MainMul(MainBase, BuyUmpMixin):
    """主裁单混特征类，UltronUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        主裁单混特征keys：['buy_deg_ang21', 'buy_price_rank252', 'buy_wave_score3', 'buy_atr_std']
        :return: ['buy_deg_ang21', 'buy_price_rank252', 'buy_wave_score3', 'buy_atr_std']
        """

        return [
            FeatureDeg().get_feature_ump_keys(ump_cls=MainMul)[-1],
            FeaturePrice().get_feature_ump_keys(ump_cls=MainMul)[-1],
            FeatureWave().get_feature_ump_keys(ump_cls=MainMul)[-1],
            FeatureAtr().get_feature_ump_keys(ump_cls=MainMul)[-1]
        ]

    def get_fiter_class(self):
        """
        主裁单混特征返回的UltronMLPd子类：UltronUmpMainMul.UmpMulFiter
        :return: UltronUmpMainMul.UmpMulFiter
        """
        return MulFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'mul_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'mul_main'



from typing import Union
import cvxpy as cp
import numpy as np
from ultron.optimize.pfopt.base import _IOptimizer

class LpOptimizer(_IOptimizer):
    def __init__(self,
                cost: np.ndarray,
                cons_matrix: np.ndarray = None,
                lower_bound: Union[float, np.ndarray] = None,
                upper_bound: Union[float, np.ndarray] = None):
        super().__init__(cost, cons_matrix, lower_bound, upper_bound)
    
    def solver(self, solver: str = "CBC"):
        x, constraints = self._prepare()

        constraints.append(sum(x) == 0.0) # 总权重之和为0.0的限制条件
        prob = cp.Problem(cp.Minimize(x @ self._cost), constraints=constraints)
        prob.solve(solver=solver)
        return x.value, prob.value, prob.status

class L1LpOptimizer(_IOptimizer):
    def __init__(self,
                 cost: np.ndarray,
                 benchmark: np.ndarray,
                 l1norm: float,
                 cons_matrix: np.ndarray = None,
                 lower_bound: Union[float, np.ndarray] = None,
                 upper_bound: Union[float, np.ndarray] = None):
        super().__init__(cost, cons_matrix, lower_bound, upper_bound)
        self._benchmark = benchmark
        self._l1norm = l1norm
    
    def solver(self, solver: str = "ECOS"):
        x, constraints = self._prepare()
        constraints.append(
            cp.pnorm(x - self._benchmark, 1) <= self._l1norm
        )
        constraints.append(cp.pnorm(x, 1) <= 2.0) # 限制多头总和为1、空头总和为-1，即相加为2.0
        constraints.append(sum(x) == 0.0) # 总权重之和为0.0的限制条件
        prob = cp.Problem(cp.Minimize(x @ self._cost), constraints=constraints)
        prob.solve(solver=solver)
        return x.value, prob.value, prob.status

    


# -*- encoding:utf-8 -*-
"""
    类基础通用模块
"""


class FreezeAttrMixin(object):
    """冻结对外设置属性混入类，设置抛异常"""

    def _freeze(self):
        """冻结属性设置接口"""
        object.__setattr__(self, "__frozen", True)

    def __setattr__(self, key, value):
        if getattr(
                self, "__frozen",
                False) and not (key in type(self).__dict__ or key == "_cache"):
            raise AttributeError(
                "You cannot add any new attribute '{key}'".format(key=key))
        object.__setattr__(self, key, value)


# -*- coding: utf-8 -*-
import numpy as np

from ultron.factor.utilities import group_mapping
from ultron.factor.utilities import simple_abssum
from ultron.factor.utilities import transform


def long_short_builder(er: np.ndarray,
                       leverage: float = 1.,
                       groups: np.ndarray = None,
                       masks: np.ndarray = None) -> np.ndarray:
    er = er.copy()

    if masks is not None:
        er[masks] = 0.
        er[~masks] = er[~masks] - er[~masks].mean()

    if er.ndim == 1:
        er = er.reshape((-1, 1))

    if groups is None:
        return er / simple_abssum(er, axis=0) * leverage
    else:
        groups = group_mapping(groups)
        return transform(groups, er, 'scale', scale=leverage)



# -*- encoding:utf-8 -*-
"""
    跳空缺口模块
"""
import numpy as np
import pandas as pd
from ultron.ump.core.helper import pd_resample
from ultron.kdutils.date import fmt_date


def calc_jump(kl_pd, jump_diff_factor=1):
    """
    通过对比交易日当月的成交量，和当月的振幅来确定交易日当日的跳空阀值，
    分别组装跳空方向，跳空能量，跳空距离等数据进入pd.DataFrame对象返回
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param jump_diff_factor: 参数通过设置jump_diff_factor来调节跳空阀值的大小，默认jump_diff_factor＝1
    :param show: 是否对结果跳空点进行可视化
    :return: pd.DataFrame对象
    """
    # 由于过程会修改金融时间序列，所以先copy一个再处理
    kl_pd = kl_pd.copy()
    # 将日change取abs变成日振幅保存在kl_pd新列abs_pct_change
    kl_pd['abs_pct_change'] = np.abs(kl_pd['p_change'])
    # 日振幅取平均做为第一层判断是否达成跳空的条件，即跳空最起码要振幅超过日振幅平均值
    change_ratio_min = kl_pd['abs_pct_change'].mean()

    # 提取月振幅volume_mean
    # TODO 做为参数可修改21d
    change_mean = pd_resample(kl_pd.abs_pct_change, '21D', how='mean')
    """
        eg: change_mean形如
        2014-07-23    0.7940
        2014-08-13    0.6536
        2014-09-03    0.8120
        2014-09-24    1.2673
        2014-10-15    1.1007
                       ...
        2016-04-13    1.2080
        2016-05-04    0.9093
        2016-05-25    0.6208
        2016-06-15    1.1831
        2016-07-06    0.6693
    """
    # 提取月成交量均值volume_mean
    volume_mean = pd_resample(kl_pd.volume, '21D', how='mean')
    """
        eg：volume_mean形如
        2014-07-23    1350679
        2014-08-13    1256093
        2014-09-03    1593358
        2014-09-24    1816544
        2014-10-15    2362897
                       ...
        2016-04-13    2341972
        2016-05-04    1633200
        2016-05-25    1372525
        2016-06-15    2071612
        2016-07-06    1136278
    """
    # 使用使用kl_pd没有resample之前的index和change_mean进行loc操作，为了把没有的index都变成nan
    change_mean = change_mean.loc[kl_pd.index.intersection(change_mean.index)]
    # 有nan之后开始填充nan
    change_mean.fillna(method='pad', inplace=True)
    # bfill再来一遍只是为了填充最前面的nan
    change_mean.fillna(method='bfill', inplace=True)
    """
        loc以及填充nan后change_mean形如：change_mean
        2014-07-23    0.7940
        2014-07-24    0.7940
        2014-07-25    0.7940
        2014-07-28    0.7940
        2014-07-29    0.7940
        2014-07-30    0.7940
        2014-07-31    0.7940
        2014-08-01    0.7940
        2014-08-04    0.7940
        2014-08-05    0.7940
                       ...
        2016-07-13    0.6693
        2016-07-14    0.6693
        2016-07-15    0.6693
        2016-07-18    0.6693
        2016-07-19    0.6693
        2016-07-20    0.6693
        2016-07-21    0.6693
        2016-07-22    0.6693
        2016-07-25    0.6693
        2016-07-26    0.6693
    """
    # 使用使用kl_pd没有resample之前的index和change_mean进行loc操作，为了把没有的index都变成nan
    volume_mean = volume_mean.loc[kl_pd.index.intersection(volume_mean.index)]
    # 有nan之后开始填充nan
    volume_mean.fillna(method='pad', inplace=True)
    # bfill再来一遍只是为了填充最前面的nan
    volume_mean.fillna(method='bfill', inplace=True)
    """
        loc以及填充nan后volume_mean形如：change_mean
        2014-07-23    1350679.0
        2014-07-24    1350679.0
        2014-07-25    1350679.0
        2014-07-28    1350679.0
        2014-07-29    1350679.0
        2014-07-30    1350679.0
        2014-07-31    1350679.0
        2014-08-01    1350679.0
        2014-08-04    1350679.0
        2014-08-05    1350679.0
                        ...
        2016-07-13    1136278.0
        2016-07-14    1136278.0
        2016-07-15    1136278.0
        2016-07-18    1136278.0
        2016-07-19    1136278.0
        2016-07-20    1136278.0
        2016-07-21    1136278.0
        2016-07-22    1136278.0
        2016-07-25    1136278.0
        2016-07-26    1136278.0
    """
    jump_pd = pd.DataFrame()

    # 迭代金融时间序列，即针对每一个交易日分析跳空
    for kl_index in np.arange(0, kl_pd.shape[0]):
        today = kl_pd.iloc[kl_index]

        if today.abs_pct_change <= change_ratio_min:
            # 第一层判断：跳空最起码要振幅超过日振幅平均值
            continue

        date = fmt_date(today.date)
        if date not in volume_mean.index:
            continue
        if today.volume <= volume_mean.loc[date]:
            # 第二层判断：跳空当日的成交量起码要超过当月平均值
            continue

        # 获取今天对应的月振幅, 做为今天判断是否跳空的价格阀值百分比
        jump_threshold = np.abs(change_mean.loc[date])
        if today.pre_close == 0 or jump_threshold == 0:
            # 只是为避免异常数据
            continue

        # 计算跳空距离阀值，即以昨天收盘为基数乘以跳空阀值比例除100得到和高开低收相同单位的价格阀值jump_diff
        # 参数通过设置jump_diff_factor来调节跳空阀值的大小，默认jump_diff_factor＝1
        jump_diff = today.pre_close * jump_threshold / 100 * jump_diff_factor
        # 第三层判断：根据向上向下跳空选择跳空计算
        if today.p_change > 0 and (today.low - today.pre_close) > jump_diff:
            # 注意向上跳空判断使用today.low，向上跳空 1
            today['jump'] = 1
            # 月振幅跳空阀值
            today['jump_threshold'] = jump_threshold
            # 跳空距离阀值
            today['jump_diff'] = jump_diff
            # 计算出跳空缺口强度
            today['jump_power'] = (today.low - today.pre_close) / jump_diff

            jump_pd = jump_pd.append(today)
        elif today.p_change < 0 and (today.pre_close - today.high) > jump_diff:
            # 注意向下跳空判断使用today.high，向下跳空 －1
            today['jump'] = -1
            # 月振幅跳空阀值
            today['jump_threshold'] = jump_threshold
            # 跳空距离阀值
            today['jump_diff'] = jump_diff
            # 计算出跳空缺口强度
            today['jump_power'] = (today.pre_close - today.high) / jump_diff
            jump_pd = jump_pd.append(today)
    return jump_pd


# -*- coding: utf-8 -*-

cdef class InverseCumulativeNormal(object):

    cdef double _average
    cdef double _sigma
    cdef int _fullAcc

    cdef double _standard_value(self, double x)
    cdef double inv(self, double x)



# -*- encoding:utf-8 -*-
"""
    拟合工具模块
"""
import math
import numpy as np
from statsmodels import api as sm, regression
from sklearn import metrics
from ultron.ump.core.helper import pd_rolling_mean


def regress_xy(x, y, mode=True, zoom=False):
    """
    使用statsmodels.regression.linear_model进行简单拟合操作，返回model和y_fit
    :param x: 可迭代序列
    :param y: 可迭代序列
    :param mode: 是否需要mode结果，在只需要y_fit且效率需要高时应设置False, 效率差异：
                 mode=False: 1000 loops, best of 3: 778 µs per loop
                 mode=True:  1000 loops, best of 3: 1.23 ms per loop
    :param zoom: 是否缩放x,y
    :param show: 是否可视化结果
    :return: model, y_fit, 如果mode=False，返回的model=None
    """
    if zoom:
        # 将y值 zoom到与x一个级别，不可用UltronScalerUtil.scaler_xy, 因为不管x > y还y > x都拿 x.max() / y.max()
        # TODO UltronScalerUtil中添加使用固定轴进行缩放的功能
        zoom_factor = x.max() / y.max()
        y = zoom_factor * y

    if mode:
        # 加常数1列
        x = sm.add_constant(x)
        model = regression.linear_model.OLS(y, x).fit()

        intercept = model.params[0]
        rad = model.params[1]
        # y = kx + b, x取x[:, 1]，因为add_constant
        y_fit = x[:, 1] * rad + intercept
    else:
        # noinspection PyCallingNonCallable
        y_fit = np.polynomial.Chebyshev.fit(x, y, 1)(x)
        model = None
    return model, y_fit


def regress_y(y, mode=True, zoom=False):
    """
    使用statsmodels.regression.linear_model进行简单拟合操作, 参数中只提供y序列，
    x使用np.arange(0, len(y))填充
    :param y: 可迭代序列
    :param mode: 是否需要mode结果，在只需要y_fit且效率需要高时应设置False, 效率差异：
             mode=False: 1000 loops, best of 3: 778 µs per loop
             mode=True:  1000 loops, best of 3: 1.23 ms per loop
    :param zoom: 是否缩放x,y
    :param show: 是否可视化结果
    :return: model, y_fit, 如果mode=False，返回的model=None
    """
    x = np.arange(0, len(y))
    return regress_xy(x, y, mode=mode, zoom=zoom)


def calc_regress_deg(y):
    """
    将y值 zoom到与x一个级别，之后再fit出弧度转成角度
    1 多个的趋势比较提供量化基础，只要同一个时间范围，就可以比较
    2 接近视觉感受到的角度
    :param y:  可迭代序列
    :param show: 是否可视化结果
    :return: deg角度float值
    """
    # 将y值 zoom到与x一个级别
    model, _ = regress_y(y, mode=True, zoom=True)
    rad = model.params[1]
    # fit出弧度转成角度
    deg = np.rad2deg(rad)
    return deg


def metrics_rmse(y, y_fit):
    """
    度量原始序列和拟合后的y_fit的RMSE：
         RMSE = np.sqrt(sum(np.square(y - y_fit)) / len(y))
    :param y: 原始可迭代序列
    :param y_fit: 拟合可迭代序列
    :return: 返回RMSE值，float
    """
    rmse = np.sqrt(metrics.mean_squared_error(y, y_fit))
    return rmse


def regress_xy_polynomial(x, y, poly=1, zoom=False):
    """
    多项式拟合, 根据参数poly决定，返回拟合后的y_fit
    :param x: 可迭代序列
    :param y: 可迭代序列
    :param poly: 几次拟合参数，int
    :param zoom: 是否对数据进行缩放
    :param show: 是否可视化显示拟合结果
    :return: y_fit
    """
    if zoom:
        # 将y值 zoom到与x一个级别，不可用UltronScalerUtil.scaler_xy, 因为不管x > y还y > x都拿 x.max() / y.max()
        zoom_factor = x.max() / y.max()
        y = zoom_factor * y

    polynomial = np.polynomial.Chebyshev.fit(x, y, poly)
    # noinspection PyCallingNonCallable
    y_fit = polynomial(x)
    return y_fit


def valid_poly(y, poly=1, zoom=False, metrics_func=metrics_rmse):
    """
    验证poly（默认＝1）次多项式拟合回归的趋势曲线是否能代表原始曲线y的走势，
    基础思路：
             1. 对原始曲线y进行窗口均线计算，窗口的大小＝ math.ceil(len(y) / 4)
             eg：
                原始y序列＝504 －> rolling_window = math.ceil(len(y) / 4) = 126
             2. 通过pd_rolling_mean计算出均线的值y_roll_mean
             3  使用metrics_func方法度量原始y值和均线y_roll_mean的距离distance_mean
             3. 通过计算regress_xy_polynomial计算多项式拟合回归的趋势曲线y_fit
             4. 使用metrics_func方法度量原始y值和拟合回归的趋势曲线y_fit的距离distance_fit
             5. 如果distance_fit <= distance_mean即代表拟合曲线可以代表原始曲线y的走势
    :param y: 原始可迭代序列
    :param poly: 几次拟合参数，int
    :param zoom: 是否对y数据进行缩放
    :param show: 是否原始曲线y，均线，以及拟合曲线可视化
    :param metrics_func: 度量始y值和均线y_roll_mean的距离和原始y值和
                         拟合回归的趋势曲线y_fit的距离的方法，默认使用metrics_rmse
    :return: 是否poly次拟合曲线可以代表原始曲线y的走势
    """
    valid = False
    x = np.arange(0, len(y))
    if zoom:
        # 将y值 zoom到与x一个级别，不可用UltronScalerUtil.scaler_xy, 因为不管x > y还y > x都拿 x.max() / y.max()
        zoom_factor = x.max() / y.max()
        y = zoom_factor * y
    # 对原始曲线y进行窗口均线计算，窗口的大小＝ math.ceil(len(y) / 4)
    rolling_window = int(math.ceil(len(y) / 4))
    # 通过pd_rolling_mean计算出均线的值y_roll_mean
    y_roll_mean = pd_rolling_mean(y, window=rolling_window, min_periods=1)
    # 使用metrics_func方法度量原始y值和均线y_roll_mean的距离distance_mean
    distance_mean = metrics_func(y, y_roll_mean, show=False)

    # 通过计算regress_xy_polynomial计算多项式拟合回归的趋势曲线y_fit, 外面做zoom了所以zoom=False
    y_fit = regress_xy_polynomial(x, y, poly=poly, zoom=False)

    # 使用metrics_func方法度量原始y值和拟合回归的趋势曲线y_fit的距离distance_fit
    distance_fit = metrics_func(y, y_fit)

    # 如果distance_fit <= distance_mean即代表拟合曲线可以代表原始曲线y的走势
    if distance_fit <= distance_mean:
        valid = True

    return valid



# -*- coding: utf-8 -*-
import sys
import unittest


class Runner:

    def __init__(self, test_cases):
        self.suite = unittest.TestSuite()
        for case in test_cases:
            tests = unittest.TestLoader().loadTestsFromTestCase(case)
            self.suite.addTests(tests)

    def run(self):
        res = unittest.TextTestRunner(verbosity=3).run(self.suite)
        if len(res.errors) >= 1 or len(res.failures) >= 1:
            sys.exit(-1)
        else:
            sys.exit(0)


# -*- coding: utf-8 -*-
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd


from ultron.factor.data.processing import factor_processing
from ultron.factor.data.standardize import standardize
from ultron.factor.data.winsorize import winsorize_normal
from ultron.optimize.constraints import Constraints
from ultron.optimize.constraints import LinearConstraints
from ultron.optimize.linearbuilder import linear_builder
from ultron.optimize.longshortbulder import long_short_builder
from ultron.optimize.meanvariancebuilder import mean_variance_builder
from ultron.factor.experimental.portfolio.meanvariancebuilder import mean_variance_builder as f_mean_variance_builder
from ultron.factor.experimental.portfolio.linearbuilder import linear_builder as f_linear_builder
from ultron.factor.analysis.simplesettle import simple_settle


def er_portfolio_analysis(er: np.ndarray,
                          industry: np.ndarray,
                          dx_return: np.ndarray,
                          constraints: Optional[Union[LinearConstraints, Constraints]] = None,
                          detail_analysis=True,
                          benchmark: Optional[np.ndarray] = None,
                          is_tradable: Optional[np.ndarray] = None,
                          method='risk_neutral',
                          **kwargs) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
    er = er.flatten()

    def create_constraints(benchmark, **kwargs):
        if 'lbound' in kwargs:
            lbound = kwargs['lbound'].copy()
            del kwargs['lbound']
        else:
            lbound = np.maximum(0., benchmark - 0.01)

        if 'ubound' in kwargs:
            ubound = kwargs['ubound'].copy()
            del kwargs['ubound']
        else:
            ubound = 0.01 + benchmark
        if is_tradable is not None:
            ubound[~is_tradable] = np.minimum(lbound, ubound)[~is_tradable]

        risk_lbound, risk_ubound = constraints.risk_targets()
        cons_exp = constraints.risk_exp
        return lbound, ubound, cons_exp, risk_lbound, risk_ubound

    if method == 'long_risk_neutral':
        lbound, ubound, cons_exp, risk_lbound, risk_ubound = create_constraints(benchmark, **kwargs)

        turn_over_target = kwargs.get('turn_over_target')
        current_position = kwargs.get('current_position')

        status, _, weights = linear_builder(er,
                                            risk_constraints=cons_exp,
                                            lbound=lbound,
                                            ubound=ubound,
                                            risk_target=(risk_lbound, risk_ubound),
                                            turn_over_target=turn_over_target,
                                            current_position=current_position)
        if status not in ("optimal", "optimal_inaccurate"):
            raise ValueError('linear programming optimizer in status: {0}'.format(status))

    elif method == 'ls' or method == 'long_short':
        weights = long_short_builder(er).flatten()
    if method == 'longshort_risk_neutral':
        lbound, ubound, cons_exp, risk_lbound, risk_ubound = create_constraints(benchmark, **kwargs)
        turn_over_target = kwargs.get('turn_over_target')
        current_position = kwargs.get('current_position')
        status, _, weights = f_linear_builder(er,
                                            risk_constraints=cons_exp,
                                            lbound=lbound,
                                            ubound=ubound,
                                            risk_target=(risk_lbound, risk_ubound),
                                            turn_over_target=turn_over_target,
                                            current_position=current_position)
        if status not in ("optimal", "optimal_inaccurate"):
            raise ValueError('linear programming optimizer in status: {0}'.format(status))

    elif method == 'longshort_mean_variance':
        lbound = kwargs['lbound']
        ubound = kwargs['ubound']
        risk_model = kwargs['risk_model']
        turn_over_target = kwargs.get('turn_over_target')
        current_position = kwargs.get('current_position')
        target_vol = kwargs.get('target_vol')
        status, _, weights = mean_variance_builder(er=er,
                                                   risk_model=risk_model,
                                                   current_position=current_position,
                                                   turn_over_target=turn_over_target,
                                                   vol_target=target_vol,
                                                   lbound=lbound,
                                                   ubound=ubound
                                                   )
        if status != 'optimal':
            raise ValueError('mean variance optimizer in status: {0}'.format(status))

    elif method == 'mv' or method == 'long_mean_variance':
        lbound, ubound, cons_exp, risk_lbound, risk_ubound = create_constraints(benchmark, **kwargs)
        risk_model = kwargs['risk_model']

        if 'lam' in kwargs:
            lam = kwargs['lam']
        else:
            lam = 1.

        status, _, weights = mean_variance_builder(er,
                                                   risk_model=risk_model,
                                                   bm=benchmark,
                                                   lbound=lbound,
                                                   ubound=ubound,
                                                   risk_exposure=cons_exp,
                                                   risk_target=(risk_lbound, risk_ubound),
                                                   lam=lam)
        if status != 'optimal':
            raise ValueError('mean variance optimizer in status: {0}'.format(status))

    else:
        raise ValueError("Unknown building type ({0})".format(method))

    if detail_analysis:
        analysis = simple_settle(weights, dx_return, industry, benchmark)
    else:
        analysis = None
    return pd.DataFrame({'weight': weights,
                         'industry': industry,
                         'er': er}), \
           analysis


# -*- encoding:utf-8 -*-
"""
    ump基础模块
"""
import weakref, functools
import numpy as np
import sklearn.preprocessing as preprocessing
from abc import ABCMeta, abstractmethod
from ultron.ump.core.fixes import six
from ultron.kdutils.file import load_pickle


class BuyUmpMixin(object):
    """
        买入ump混入, 与BuyFeatureMixin不同，每一个具体的ump只能属于一个ump类别
        即不是BuyUmpMixin就应该是SellUmpMixin
    """
    _ump_type_prefix = 'buy_'


class SellUmpMixin(object):
    """
        卖出ump混入, 与SellFeatureMixin不同，每一个具体的ump只能属于一个ump类别
        即不是BuyUmpMixin就应该是SellUmpMixin
    """
    _ump_type_prefix = 'sell_'


class UmpDict(dict):
    """Several built-in types such as list and dict do not directly support weak references
     but can add support through subclassing:"""
    pass


class CachedUmpManager:
    """ump拦截缓存实体，分别在主裁和边裁类中"""
    """不对外开发的设置，仅针对源代码修改，默认使用dict不使用WeakValueDictionary"""
    s_use_weak = False

    def __init__(self):
        """初始化_cache本体，根据s_use_weak决定使用WeakValueDictionary或者dict"""
        self._cache = weakref.WeakValueDictionary(
        ) if CachedUmpManager.s_use_weak else dict()

    def get_ump(self, ump):
        """
        主要在具体裁判类的predict方法中获取裁判本体使用，如果
        不在_cache中load_pickle，否则使用catch中的ump返回
        :param ump: 具体裁判对象，UltronUmpBase对象
        :return: 每个裁判需要的决策数据，每类裁判主体的数据形式不同，且使用方法不同
                 eg：主裁中的使用
                    def predict(self, x, need_hit_cnt=1):
                        dump_clf_with_ind = UltronUmpMainBase.dump_clf_manager.get_ump(self)
                        count_hit = 0
                        for clf, ind in dump_clf_with_ind.values():
                            ss = clf.predict(x)
                            if ss == ind:
                                count_hit += 1
                                if need_hit_cnt == count_hit:
                                    return 1
                        return 0
        """
        # dump_file_fn是每一个具体裁判需要复写的方法，声明自己缓存的存放路径
        name = ump.dump_file_fn()  #fix me
        if name not in self._cache:
            # 不在缓存字典中load_pickle
            dump_clf = load_pickle(name)
            if dump_clf is None:
                # 没有对ump进行训练，就直接进行拦截了，抛异常
                raise RuntimeError(
                    '{}: you must first fit orders, or {} is not exist!!'.
                    format(ump.__class__.__name__, name))
            if CachedUmpManager.s_use_weak:
                # 如果使用WeakValueDictionary模式，需要进一步使用UmpDict包一层
                dump_clf = UmpDict(**dump_clf)
            self._cache[name] = dump_clf
        else:
            # 有缓存直接拿缓存
            dump_clf = self._cache[name]
        return dump_clf

    def clear(self):
        """清除缓存中所有cache ump"""
        self._cache.clear()


def ump_main_make_xy(func):
    """
    主裁中对应fiter class中make_xy的装饰器，
    使用eg：详阅UltronUmpMainDeg或其它主裁子类实现

        class UltronUmpMainDeg(UltronUmpMainBase, BuyUmpMixin):
            class UmpDegFiter(UltronMLPd):
                @ump_main_make_xy
                def make_xy(self, **kwarg):
                    regex = 'result|{}'.format(
                        '|'.join(UltronFeatureDeg().get_feature_ump_keys(ump_cls=UltronUmpMainDeg)))
                    # noinspection PyUnresolvedReferences
                    deg_df = self.order_has_ret.filter(regex=regex)
                    return deg_df
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        kwargs中必须的参数为：kwargs['orders_pd']，在UltronUmpMainBase初始化__init__中
        将orders_pd=self.orders_pd，eg：
            self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)

        kwargs中可选的参数为：kwargs['scaler']，bool类型，默认不传递为false，控制在
        make_xy中返回的特征矩阵数据是否进行标准化处理
        """
        if kwargs is None or 'orders_pd' not in kwargs:
            raise ValueError('kwarg is None or not kwarg.has_key orders_pd')

        orders_pd = kwargs['orders_pd']
        # 从orders_pd中筛选有交易结果形成order_has_ret
        order_has_ret = orders_pd[orders_pd['result'] != 0]

        # 之前的交易结果－1为loss，1为win，0为keep状态，order_has_ret没有keep状态，所以转换为loss：0，win：1
        # noinspection PyTypeChecker
        order_has_ret['result'] = np.where(order_has_ret['result'] == -1, 0, 1)
        self.order_has_ret = order_has_ret
        # 通过被装饰的make_xy方法，筛选具体裁判需要的特征形成特征矩阵ump_df
        ump_df = func(self, *args, **kwargs)

        if 'scaler' in kwargs and kwargs['scaler'] is True:
            # 控制在make_xy中返回的特征矩阵数据是否进行标准化处理
            scaler = preprocessing.StandardScaler()
            for col in ump_df.columns[1:]:
                ump_df[col] = scaler.fit_transform(ump_df[col].values.reshape(
                    -1, 1))
        # 转换为matrix，形成x，y
        ump_np = ump_df.values  #ump_df.as_matrix()
        self.y = ump_np[:, 0]
        self.x = ump_np[:, 1:]
        # 将pd.DataFrame对象ump_df也保留一份
        self.df = ump_df
        self.np = ump_np

    return wrapper


def ump_edge_make_xy(func):
    """
    边裁中对应fiter class中make_xy的装饰器，
    使用eg：详阅UltronUmpEdgeDeg或其它边裁子类实现

        class UltronUmpEdgeDeg(UltronUmpEdgeBase, BuyUmpMixin):
            class UmpDegFiter(UltronMLPd):
                @ump_edge_make_xy
                def make_xy(self, **kwarg):
                    filter_list = ['profit', 'profit_cg']
                    cols = UltronFeatureDeg().get_feature_ump_keys(ump_cls=UltronUmpEdgeDeg)
                    filter_list.extend(cols)
                    # noinspection PyUnresolvedReferences
                    deg_df = self.order_has_ret.filter(filter_list)
                    return deg_df
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        kwargs中必须的参数为：kwargs['orders_pd']，在UltronUmpEdgeBase初始化__init__中
        将orders_pd=self.orders_pd，eg：
            self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
        """
        if kwargs is None or 'orders_pd' not in kwargs:
            raise ValueError('kwarg is None or not kwarg.has_key orders_pd')

        orders_pd = kwargs['orders_pd']
        # 从orders_pd中筛选有交易结果形成orders_pd_tmp
        orders_pd_tmp = orders_pd[orders_pd['result'] != 0]
        # 从orders_pd_tmp进行二次筛选必须profit != 0
        order_has_ret = orders_pd_tmp[orders_pd_tmp['profit'] != 0]
        self.order_has_ret = order_has_ret
        # 通过被装饰的make_xy方法，筛选具体裁判需要的特征形成特征矩阵ump_df
        ump_df = func(self, *args, **kwargs)
        # 转换为matrix，形成x，y
        ump_np = ump_df.values  #ump_df.as_matrix()
        # 边裁特征中filter_list = ['profit', 'profit_cg']都设定为y
        self.y = ump_np[:, :2]
        self.x = ump_np[:, 2:]
        # 将pd.DataFrame对象ump_df也保留一份
        self.df = ump_df
        self.np = ump_np

    return wrapper


class Base(six.with_metaclass(ABCMeta, object)):
    """ump拦截缓存，在UltronUmpBase类中"""
    dump_clf_manager = CachedUmpManager()

    @abstractmethod
    def get_fiter_class(self):
        """abstractmethod子类必须实现，声明具体子类裁判使用的筛选特征形成特征的类"""
        pass

    @abstractmethod
    def get_predict_col(self):
        """abstractmethod子类必须实现，获取具体子类裁判需要的特征keys"""
        pass

    @classmethod
    def is_buy_ump(cls):
        """
        返回裁判本身是否是买入拦截裁判，类方法
        :return: bool，是否是买入拦截裁判
        """
        return getattr(cls, "_ump_type_prefix") == 'buy_'

    @classmethod
    @abstractmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法，abstractmethod子类必须实现
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        pass

    def __str__(self):
        """打印对象显示：class name, is_buy_ump,  predict_col"""
        return '{}: is_buy_ump:{} predict_col:{}'.format(
            self.__class__.__name__, self.__class__.is_buy_ump(),
            self.get_predict_col())

    __repr__ = __str__



# -*- coding: utf-8 -*-
import re
cimport cython
from libc.math cimport floor
from ultron.tradingday.Enums._TimeUnits cimport TimeUnits


_unitPattern = re.compile('[BbDdMmWwYy]{1}')
_numberPattern = re.compile('[-+]*[0-9]+')

_unitsDict = {'d': TimeUnits.Days,
              'b': TimeUnits.BDays,
              'w': TimeUnits.Weeks,
              'm': TimeUnits.Months,
              'y': TimeUnits.Years}


cdef class Period(object):

    def __init__(self, str reprStr=None, int length=0, int units=0):

        cdef int n
        cdef str unitsStr

        if reprStr:
            unitsPos = _unitPattern.search(reprStr)
            numPos = _numberPattern.search(reprStr)
            unitsStr = reprStr[unitsPos.start():unitsPos.end()].lower()
            n = int(reprStr[numPos.start():numPos.end()])
            self._length = n
            self._units = int(_unitsDict[unitsStr])
        else:
            self._length = length
            self._units = units

    cpdef int length(self):
        return self._length

    cpdef int units(self):
        return self._units

    cpdef Period normalize(self):
        cdef int length = self.length()
        cdef int units = self.units()
        if length != 0:
            if units == TimeUnits.BDays or units == TimeUnits.Weeks or units == TimeUnits.Years:
                return Period(length=length, units=units)
            elif units == TimeUnits.Months:
                if length % 12 == 0:
                    length //= 12
                    units = TimeUnits.Years
                return Period(length=length, units=units)
            elif units == TimeUnits.Days:
                if length % 7 == 0:
                    length //= 7
                    units = TimeUnits.Weeks
                return Period(length=length, units=units)
            else:
                pass
                #ensureRaise(TypeError, "wrong period units type {0}".format(units))

    def __div__(self, int n):

        cdef int resunits = self.units()
        cdef int reslength = self.length()

        if reslength % n == 0:
            reslength /= n
        else:
            if resunits == TimeUnits.Years:
                reslength *= 12
                resunits = TimeUnits.Months
            elif resunits == TimeUnits.Weeks:
                reslength *= 7
                resunits = TimeUnits.Days

            #require(reslength % n == 0, ValueError, "{0} cannot be divided by {1:d}".format(self, n))

            reslength //= n

        return Period(length=reslength, units=resunits)


    # only work for python 3
    def __truediv__(self, int n):

        cdef int resunits = self.units()
        cdef int reslength = self.length()

        if reslength % n == 0:
            reslength /= n
        else:
            if resunits == TimeUnits.Years:
                reslength *= 12
                resunits = TimeUnits.Months
            elif resunits == TimeUnits.Weeks:
                reslength *= 7
                resunits = TimeUnits.Days

            #require(reslength % n == 0, ValueError, "{0} cannot be divided by {1:d}".format(self, n))
            reslength //= n

        return Period(length=reslength, units=resunits)

    def __add__(self, p2):

        cdef int reslength = self.length()
        cdef int resunits = self.units()
        cdef int p2length = p2.length()
        cdef int p2units = p2.units()

        if reslength == 0:
            return Period(length=p2length, units=p2units)
        elif resunits == p2units:
            reslength += p2length
            return Period(length=reslength, units=resunits)
        else:
            if resunits == TimeUnits.Years:
                if p2units == TimeUnits.Months:
                    resunits = TimeUnits.Months
                    reslength = reslength * 12 + p2length
                elif p2units == TimeUnits.Weeks or p2units == TimeUnits.Days or p2units == TimeUnits.BDays:
                    pass
                    #require(p2length == 0, ValueError, "impossible addition between {0} and {1}".format(self, p2))
                else:
                    raise ValueError("unknown time unit ({0:d})".format(p2units))
                return Period(length=reslength, units=resunits)
            elif resunits == TimeUnits.Months:
                if p2units == TimeUnits.Years:
                    reslength += 12 * p2length
                elif p2units == TimeUnits.Weeks or p2units == TimeUnits.Days or p2units == TimeUnits.BDays:
                    pass
                    #require(p2length == 0, ValueError, "impossible addition between {0} and {1}".format(self, p2))
                else:
                    raise ValueError("unknown time unit ({0:d})".format(p2units))
                return Period(length=reslength, units=resunits)
            elif resunits == TimeUnits.Weeks:
                if p2units == TimeUnits.Days:
                    resunits = TimeUnits.Days
                    reslength = reslength * 7 + p2length
                elif p2units == TimeUnits.Years or p2units == TimeUnits.Months or p2units == TimeUnits.BDays:
                    pass
                    #require(p2length == 0, ValueError, "impossible addition between {0} and {1}".format(self, p2))
                else:
                    raise ValueError("unknown time unit ({0:d})".format(p2units))
                return Period(length=reslength, units=resunits)
            elif resunits == TimeUnits.Days:
                if p2units == TimeUnits.Weeks:
                    reslength += 7 * p2length
                elif p2units == TimeUnits.Years or p2units == TimeUnits.Months or p2units == TimeUnits.BDays:
                    pass
                    #require(p2length == 0, ValueError, "impossible addition between {0} and {1}".format(self, p2))
                else:
                    raise ValueError("unknown time unit ({0:d})".format(p2units))
                return Period(length=reslength, units=resunits)
            elif resunits == TimeUnits.BDays:
                if p2units == TimeUnits.Years or p2units == TimeUnits.Months or p2units == TimeUnits.Weeks or p2units == TimeUnits.Days:
                    pass
                    #require(p2length == 0, ValueError, "impossible addition between {0} and {1}".format(self, p2))
                else:
                    raise ValueError("unknown time unit ({0:d})".format(p2units))
                return Period(length=reslength, units=resunits)

    def __neg__(self):
        return Period(length=-self.length(), units=self.units())

    def __sub__(self, p2):
        return self + (-p2)

    def __str__(self):
        cdef str out = ""
        cdef int n = self.length()
        cdef int m = 0
        cdef int units = self.units()

        if units == TimeUnits.Days:
            if n >= 7:
                m = int(floor(n / 7,))
                out = '{0}W'.format(m)
                n %= 7
            if n != 0 or m == 0:
                return '{0}{1}D'.format(out, n)
            else:
                return out
        elif units == TimeUnits.Weeks:
            return '{0}W'.format(n)
        elif units == TimeUnits.Months:
            if n >= 12:
                m = int(floor(n / 12.))
                out = '{0}Y'.format(m)
                n %= 12
            if n != 0 or m == 0:
                return '{0}{1}M'.format(out, n)
            else:
                return out
        elif units == TimeUnits.Years:
            return '{0}Y'.format(n)
        elif units == TimeUnits.BDays:
            return '{0}B'.format(n)

    def __richcmp__(self, right, int op):
        if op == 0:
            return _lt_cmp(self, right)
        elif op == 2:
            return not (_lt_cmp(self, right) or _lt_cmp(right, self))
        elif op == 3:
            return _lt_cmp(self, right) or _lt_cmp(right, self)
        elif op == 4:
            return _lt_cmp(right, self)

    def __deepcopy__(self, memo):
        return Period(length=self.length(), units=self.units())

    def __reduce__(self):
        d = {}

        return Period, (None, self.length(), self.units()), d

    def __setstate__(self, state):
        pass


cpdef check_period(p):
    if isinstance(p, Period):
        return p
    else:
        return Period(p)


# implementation detail

cdef bint _lt_cmp(Period p1, Period p2) except -1:

    cdef tuple p1lim
    cdef tuple p2lim
    cdef int p1length = p1.length()
    cdef int p1units = p1.units()
    cdef int p2length = p2.length()
    cdef int p2units = p2.units()

    if p1length == 0:
        return p2length > 0

    if p2length == 0:
        return p1length < 0

    # exact comparisons
    if p1units == p2units:
        return p1length < p2length
    elif p1units == TimeUnits.Months and p2units == TimeUnits.Years:
        return p1length < (p2length * 12)
    elif p1units == TimeUnits.Years and p2units == TimeUnits.Months:
        return (p1length * 12) < p2length
    elif p1units == TimeUnits.Days and p2units == TimeUnits.Weeks:
        return p1length < (p2length * 7)
    elif p1units == TimeUnits.Weeks and p2units == TimeUnits.Days:
        return (p1length * 7) < p2length

    # inexact comparisons (handled by converting to days and using limits)

    p1lim = _daysMinMax(p1)
    p2lim = _daysMinMax(p2)

    if p1lim[1] < p2lim[0]:
        return True
    elif p1lim[0] >= p2lim[1]:
        return False
    else:
        pass
        #ensureRaise(ValueError, "undecidable comparison between {0} and {1}".format(p1, p2))


cdef tuple _daysMinMax(Period p):

    cdef int units = p.units()
    cdef int length = p.length()

    if units == TimeUnits.Days:
        return length, length
    elif units == TimeUnits.Weeks:
        return 7 * length, 7 * length
    elif units == TimeUnits.Months:
        return 28 * length, 31 * length
    elif units == TimeUnits.Years:
        return 365 * length, 366 * length
    elif units == TimeUnits.BDays:
        pass
        #ensureRaise(ValueError, "Business days unit has not min max days")


# -*- coding: utf-8 -*-

cdef public enum BizDayConventions:
    Following = 0
    ModifiedFollowing = 1
    Preceding = 2
    ModifiedPreceding = 3
    Unadjusted = 4
    HalfMonthModifiedFollowing = 5
    Nearest = 6